local CP_UI, UI = ...
----------------------------------
_G[CP_UI] = UI
----------------------------------------------------------------
local 	assert, format, pairs, ipairs, next, select, type, unpack, wipe, tconcat = 
		assert, format, pairs, ipairs, next, select, type, unpack, wipe, table.concat
local 	parsePoints, parseConstruct, calcRelativeRegion,
		runRegionFunction, runRegionFunctionSafe,
		global_mixin, mixin, addSubTablem, throwError
----------------------------------------------------------------
local ARGUMENTS, ERROR, ERROR_CODES, REGION, SETUP, IGNORE_SETUP, IGNORE_VALUE
local POINTS_TO_PARSE, CONSTRUCT_TO_PARSE = {}, {}
----------------------------------------------------------------

--- Create a new frame.
-- @param 	object 	: Type of object to be created. Frame or inherited therefrom. 
-- @param 	name 	: Name of the object to be created.
-- @param 	parent 	: Parent of the object.
-- @param 	templates : Inherited templates.
-- @param 	wireframe : Table consisting of additional regions to be created.
-- @return 	frame 	: Returns the created object.
function UI:CreateFrame(object, name, parent, templates, wireframe, recursiveCall)
	----------------------------------
	if templates and templates:match('Secure') then
		assert(not InCombatLockdown(), 'Cannot create secure frame in combat!')
	end
	----------------------------------
	local frame = CreateFrame(object, name, parent, templates)
	----------------------------------
	if wireframe then
		self:BuildWireframe(frame, wireframe, true)
	end
	----------------------------------
	if not recursiveCall then
		parsePoints()
		parseConstruct()
	end
	return frame
end

--- Build wireframe on existing frame.
-- @param 	frame 	: Parent of the wireframe to-be.
-- @param 	wireframe : Wireframe to be constructed.
-- @return 	frame 	: Returns the altered frame.
function UI:BuildWireframe(frame, wireframe, recursiveCall)
	assert(type(wireframe) == 'table', throwError('Wireframe', frame:GetName(), ERROR_CODES.WIREFRAME))
	for key, config in pairs(wireframe) do
		local isLoop = config.Repeat
		for i = 1, ( config.Repeat or 1 ) do
			local key = ( isLoop and key..i ) or key
			local region = frame[key]
			if not region then
				----------------------------------
				if config.Setup then -- Assert type exists if setup table exists.
					assert(config.Type, throwError(key, name, ERROR_CODES.REGION))
				end
				----------------------------------
				-- Region type has special constructor.
				if config.Type then
					if REGION[config.Type] then
						region = REGION[config.Type](frame, key, config.Setup or config.Values)
					-- Region already exists.
					elseif type(config.Type) == 'table' then
						region = REGION.Existing(frame, key, config.Type)
					-- Region should be a type of frame.
					elseif type(config.Type) == 'string' then
						region = self:CreateFrame(config.Type, '$parent'..key, frame, config.Setup and tconcat(config.Setup, ', '), config[1], true)
						-- Children have already been added recursively, remove the wireframe.
						config[1] = nil
					end
				else -- Assume this is a data table.
					region = config
				end

				frame[key] = region
			end

			if isLoop and region.SetID then
				region:SetID(i)
			end

			-- apply preset (if present) before everything else
			if config.Preset then
				runRegionFunctionSafe(region, 'Preset', config.Preset)
			end

			-- mixin before running the rest of the region func stack,
			-- since mixed in functions may be called from the wireframe
			if config.Mixin then
				runRegionFunctionSafe(region, 'Mixin', config.Mixin)
			end

			for sType, sData in pairs(config) do
				runRegionFunction(region, sType, sData)
			end
		end
	end
	-- parse if function is explicitly called
	if not recursiveCall then
		parsePoints()
		parseConstruct()
	end
	return frame
end

function runRegionFunctionSafe(region, sType, sData)
	if sData == 'nil' then
		sData = nil
	end
	local func = SETUP[sType] or region[sType]
	if type(func) == 'function' then
		-- if non-associative table, just unpack it.
		if ( type(sData) == 'table' and #sData ~= 0 and sType ~= 1) then
			return func(region, unpack(sData))
		else
			return func(region, sData)
		end
	elseif not IGNORE_VALUE[sType] then
		region[sType] = sData
	end
end

function runRegionFunction(region, sType, sData)
	if not IGNORE_SETUP[sType] then
		runRegionFunctionSafe(region, sType, sData)
	end
end


----------------------------------
function calcRelativeRegion(region, regionString)
	local relativeRegion
	for key in regionString:gmatch('%a+') do
		if key == 'parent' then
			relativeRegion = relativeRegion and relativeRegion:GetParent() or region:GetParent()
		elseif relativeRegion then
			relativeRegion = relativeRegion[key]
		else
			relativeRegion = region[key]
		end
	end
	return relativeRegion
end

function parsePoints()
	for _, setup in pairs(POINTS_TO_PARSE) do
		local numArgs = #setup
		if numArgs == 2 then
			local region, point = unpack(setup)
			region:SetPoint(point)
		elseif numArgs == 4 then
			local region, point, xOffset, yOffset = unpack(setup)
			region:SetPoint(point, xOffset, yOffset)
		elseif numArgs == 6 then
			local region, point, relativeRegion, relativePoint, xOffset, yOffset = unpack(setup)
			if type(relativeRegion) == 'table' then
				region:SetPoint(point, relativeRegion, relativePoint, xOffset, yOffset)
			elseif type(relativeRegion == 'string') then
				region:SetPoint(point, calcRelativeRegion(region, relativeRegion), relativePoint, xOffset, yOffset)
			end
		end
	end
	wipe(POINTS_TO_PARSE)
end

function parseConstruct()
	for _, setup in pairs(CONSTRUCT_TO_PARSE) do
		local region, constructor = unpack(setup)
		assert(type(constructor) == 'function', throwError('Constructor', region:GetName(), ERROR_CODES.CONSTRUCTOR))
		constructor(region)
	end
	wipe(CONSTRUCT_TO_PARSE)
end


----------------------------------
----------------------------------

function addSubTable(tbl, ...) tbl[#tbl + 1] = {...} end

function global_mixin(object, ...)
	for i = 1, select('#', ...) do
		local mixin = select(i, ...)
		for k, v in pairs(mixin) do
			object[k] = v
		end
	end

	return object
end

function mixin(t, ...)
	t = global_mixin(t, ...)
	if t.HasScript then
		for k, v in pairs(t) do
			if t:HasScript(k) then
				if t:GetScript(k) then
					t:HookScript(k, v)
				else
					t:SetScript(k, v)
				end
			end
		end
	end
end

----------------------------------
ARGUMENTS = {
----------------------------------
	COLOR 	= 'red, green, blue [, alpha]',
	COORDS 	= '(left, right, top, bottom) or (ULx, ULy, LLx, LLy, URx, URy, LRx, LRy)',
	SIZE 	= 'width, height',
	POINT 	= '(point) or (point, xOffset, yOffset) or (point, relativePoint, xOffset, yOffset) or (point, relativeRegion, relativePoint, xOffset, yOffset)',
	WIREFRAME = '(frame, wireframe); wireframe = { child1 = {}, child2 = {}, ..., childN = {} }',
}---------------------------------

----------------------------------
ERROR = '%s in %s %s.'
----------------------------------
ERROR_CODES = {
----------------------------------
	COLOR 	= 'has incorrect color values. Arguments: '..ARGUMENTS.COLOR,
	COORDS	= 'has incorrect texture coordinates. Arguments: '..ARGUMENTS.COORDS,
	POINT	= 'has incorrect point details. Arguments: '..ARGUMENTS.POINT,
	SIZE	= 'has incorrect size values. Arguments: '..ARGUMENTS.SIZE,
	------------------------------
	MULTI_FUNC 	= 'does not exist',
	MULTI_TABLE = 'has an invalid loop table',
	REGION 	= 'missing region type',
	SETUP 	= 'missing setup table',
	CONSTRUCTOR = 'is invalid. Constructor must be a function.',
	WIREFRAME = 'is invalid. Wireframe should be a nested table. Arguments: '..ARGUMENTS.WIREFRAME,
}---------------------------------

function throwError(key, name, code) return format(ERROR, key, name or 'unnamed region', code) end

----------------------------------
REGION = {
----------------------------------
	AnimationGroup = function(parent, key, setup) return parent:CreateAnimationGroup('$parent'..key, setup and unpack(setup)) end,
	Animation = function(parent, key, setup) return parent:CreateAnimation(setup, '$parent'..key) end,
	FontString = function(parent, key, setup) return parent:CreateFontString('$parent'..key, setup and unpack(setup)) end,
	Texture = function(parent, key, setup) return parent:CreateTexture('$parent'..key, setup and unpack(setup)) end,
	---
	ScrollFrame = function(parent, key, setup)
		local frame = CreateFrame('ScrollFrame', '$parent'..key, parent)
		local child = setup or CreateFrame('Frame', '$parentChild', frame)
		frame.Child = child
		child:SetParent(frame)
		child:SetAllPoints()
		frame:SetScrollChild(child)
		return frame
	end,
	---
	Table = function(parent, key, setup) parent[key] = setup or {} return parent[key] end,
	---
	Existing = function(parent, key, region)
		_G[parent:GetName()..key] = region
		region:SetParent(parent)
		return region
	end
}---------------------------------

----------------------------------
SETUP = {
----------------------------------
	[1] = function(frame, wireframe) UI:BuildWireframe(frame, wireframe, true) end,
	--- Texture
	Atlas 	= function(texture, ...) texture:SetAtlas(...) end,
	Blend 	= function(texture, ...) texture:SetBlendMode(...) end,
	Coords 	= function(texture, ...) texture:SetTexCoord(...) end,
	Gradient = function(texture, ...) texture:SetGradientAlpha(...) end,
	Texture = function(texture, ...) texture:SetTexture(...) end,
	--- FontString
	AlignH 	= function(fontString, ...) fontString:SetJustifyH(...) end,
	AlignV 	= function(fontString, ...) fontString:SetJustifyV(...) end,
	Color 	= function(fontString, ...) fontString:SetTextColor(...) end,
	Font 	= function(fontString, ...) fontString:SetFont(...) end,
	FontH 	= function(fontString, ...) fontString:SetHeight(...) end,
	Text 	= function(fontString, ...) fontString:SetText(...) end,
	--- LayeredRegion
	Layer 	= function(region, ...) region:SetDrawLayer(...) end,
	Vertex 	= function(region, ...) region:SetVertexColor(...) end,
	--- Frame
	Attrib 	= function(frame, attributes) for k, v in pairs(attributes) do frame:SetAttribute(k, v) end end,
	Backdrop = function(frame, backdrop) frame:SetBackdrop(backdrop) end,
	Background = function(frame, backdrop) UI.Media:SetBackdrop(frame, backdrop) end,
	Hooks 	= function(frame, scripts) for k, v in pairs(scripts) do frame:HookScript(k, v) end end,
	Level 	= function(frame, ...) frame:SetFrameLevel(...) end,
	Strata 	= function(frame, ...) frame:SetFrameStrata(...) end,
	Scripts = function(frame, scripts) for k, v in pairs(scripts) do frame:SetScript(k, v) end end,
	--- Region
	Alpha 	= function(region, ...) region:SetAlpha(...) end,
	Clear 	= function(region) region:ClearAllPoints() end,
	Fill 	= function(region, target) region:SetAllPoints(target ~= true and target) end,
	Height 	= function(region, ...) region:SetHeight(...) end,
	Hide 	= function(region, ...) region:Hide() end,
	Preset 	= function(region, ...) UI:ApplyTemplate(region, ...) end,
	Probe 	= function(region, ...) region.probe = UI:CreateProbe(region, ...) end,
	Show 	= function(region, ...) region:Show() end,
	Size 	= function(region, ...) region:SetSize(...) end,
	Scale 	= function(region, ...) region:SetScale(...) end,
	Width 	= function(region, ...) region:SetWidth(...) end,
	--- Constructor
	OnLoad 	= function(region, ...) addSubTable(CONSTRUCT_TO_PARSE, region, ...) end,
	--- Points
	Point 	= function(region, ...) addSubTable(POINTS_TO_PARSE, region, ...) end,
	Points  = function(region, ...) for _, point in pairs({...}) do addSubTable(POINTS_TO_PARSE, region, unpack(point)) end end,
	-- Mixin 
	Mixin 	= function(region, ...) UI:ApplyMixin(region, nil, ...) end,
	MixinNormal = function(region, ...) UI:ApplyMixin(region, UI.Utils.MixinNormal, ...) end,
	--- Multiple runs
	Multiple 	= function(region, multiTable)
		for k, v in pairs(multiTable) do
			assert(region[k], throwError(k, region:GetName(), ERROR_CODES.MULTI_FUNC))
			assert(type(v) == 'table', throwError(k, region:GetName(), ERROR_CODES.MULTI_TABLE))
			for _, args in pairs(v) do
				runRegionFunction(region, k, args)
			end
		end
	end,
}---------------------------------

----------------------------------
IGNORE_SETUP = {
----------------------------------
	Preset = true,
	Repeat = true,
	Mixin = true,
}---------------------------------

----------------------------------
IGNORE_VALUE = {
----------------------------------
	Type = true,
	Setup = true,
	Values = true,
}---------------------------------

-- Allow other entities to access these objects directly
UI.Functions = SETUP
UI.Region = REGION
-- Shortcut for mixin with scripts, since it's expected to be used frequently.
UI.Mixin = SETUP.Mixin


